/*
 * px500.h -- definitions for the px500 frame grabber
 *
 *********/

#ifndef _LINUX_PX500_H_
#define _LINUX_PX500_H_

#include <linux/ioctl.h>
#include <linux/types.h>
#include <linux/delay.h>

/* version dependencies have been confined to a separate file */

#include "sysdep-2.1.h"


/*
 * Macros to help debugging
 */

#undef PDEBUG             /* undef it, just in case */
#undef DEBUG_CODE
#ifdef PX_DEBUG
#  ifdef __KERNEL__
     /* This one if debugging is on, and kernel space */
#    define PDEBUG(fmt, args...) printk( KERN_DEBUG "pxdrv: " fmt, ## args)
#    define DEBUG_CODE(code) code
#  else
     /* This one for user space */
#    define PDEBUG(fmt, args...) fprintf(stderr, fmt, ## args)
#  endif
#else
#  define PDEBUG(fmt, args...) /* not debugging: nothing */
#  define DEBUG_CODE(code)
#endif

#undef PDEBUGG
#define PDEBUGG(fmt, args...) /* nothing: it's a placeholder */

/*
 * First of all, declare symbols to access the various registers.
 * They can be used by user-space programs via ioctl().
 * The macros with a leading double underscore should only be used
 * via specific functions, as the value is cached in the Dev struct
 */

#define PX_REG_DMA0    0x00
#define PX_REG_DMA1    0x0c
#define PX_REG_HOLE0   0x18
#define PX_REG_HOLE1   0x24
#define PX_REG_MODE0   0x30
#define PX_REG_MODE1   0x34
#define PX_REG_FIFOTRG 0x38
#define PX_REG_SETTO3  0x3c
#define __PX_REG_GRABCTL 0x40
#define     PX_REG_GRAB_ON   0x000080F3
#define     PX_REG_GRAB_OFF  0x000080F0
#define     PX_REG_GRAB_NONE 0x00008040

#define PX_REG_WAITCNT 0x44
#define PX_REG_INTRCTL 0x48
#define PX_REG_MASK0   0x4c
#define PX_REG_MASK1   0x50
#define PX_REG_MASKL   0x54
#define PX_REG_FIFOCTL 0x58
#define PX_REG_SERTIME 0x5c
#define PX_REG_SERCTL  0x60

#define PX_REG_SERDATA 0x64
#define PX_REG_DMAEND0 0x8c
#define PX_REG_DMAEND1 0x90


/*
 * These are the ports
 */
#define __PX_PORT_CONTROL            0x30
#define    PX_BIT_DUMP_CACHE   0x01
#define    PX_BIT_LOAD_CACHE   0x02
#define    PX_BIT_TRIGGER_W    0x04
#define    PX_BIT_INPUT_0      0x08
#define    PX_BIT_INPUT_1      0x10
#define    PX_BIT_TRIGGER_ACK  0x20
#define    PX_BIT_IRQ_ENABLE   0x40
#define    PX_BIT_IRQ_ACK      0x80
#define PX_PORT_MICRO              0x31
#define __PX_PORT_TRIGGER          0x32
#define    PX_BIT_TRIGGER_POL  0x01
#define    PX_BIT_TRIGGER_EDGE 0x02
#define    PX_BIT_TRIGGER_DEBN 0x03
#define    PX_BIT_INIT_STATUS  0x08
#define    PX_BIT_DECIMATE0    0x10
#define    PX_BIT_DECIMATE1    0x20
#define    PX_BIT_DECIMATE2    0x40
#define PX_PORT_STATUS             0x33
#define    PX_BIT_IRQ_STATUS   0x01
#define    PX_BIT_GREATER      0x02
#define    PX_BIT_FIELD        0x04
#define    PX_BIT_TRIG_STATUS  0x08
#define    PX_BIT_CACHED       0x10
#define    PX_BIT_EQUAL        0x20
#define    PX_BIT_HAVE_VIDEO   0x40
#define    PX_BIT_MP_ACK       0x80
#define PX_PORT_ID                 0x34
#define PX_PORT_COMPARE            0x34
#define PX_PORT_ADDRESS            0x35
#define PX_PORT_DATA               0x36
#define PX_PORT_LEVEL              0x37

#define PX_CMD_NOP          0x00
#define PX_CMD_STARTLINE    0x0a
#define PX_CMD_ENDLINE      0x0b
#define PX_CMD_REVISION     0x11
#define PX_CMD_STATUS       0x12
#define    PX_BIT_PAL           0x01
#define    PX_BIT_NORAM         0x02
#define PX_CMD_PROTECTION   0x13
#define PX_CMD_RD_STARTLINE 0x1a
#define PX_CMD_RD_ENDLINE   0x1b

/*
 * Ioctl definitions
 */

#define PX_IOC_MAGIC  'p' /* Use 'p' as magic number */

/* Reset cmds */
#define PX_IOCRESET         _IO(PX_IOC_MAGIC, 0)
#define PX_IOCHARDRESET     _IO(PX_IOC_MAGIC, 1) /* debugging tool */

/* Get the whole structure to see it */

#define PX_IOCGDEV          _IOR(PX_IOC_MAGIC, 2, Px_Dev)

/* Reading values */
#define PX_IOCHAVECACHE     _IOR(PX_IOC_MAGIC, 3, char)

/* Modifying parameters */
#define PX_IOCCMD          _IOWR(PX_IOC_MAGIC, 4, Px_Register)
#define PX_IOCSREG          _IOW(PX_IOC_MAGIC, 5, Px_Register)
#define PX_IOCGREG         _IOWR(PX_IOC_MAGIC, 6, Px_Register)
#define PX_IOCSPORT         _IOW(PX_IOC_MAGIC, 7, Px_Register)
#define PX_IOCGPORT        _IOWR(PX_IOC_MAGIC, 8, Px_Register)
#define PX_IOCSLUT          _IOW(PX_IOC_MAGIC, 9, Px_Lut)
#define PX_IOCGLUT          _IOR(PX_IOC_MAGIC,10, Px_Lut)
#define PX_IOCSGAIN         _IOW(PX_IOC_MAGIC,11, Px_Gain)
#define PX_IOCGGAIN         _IOR(PX_IOC_MAGIC,12, Px_Gain)

#define PX_IOCSSIZE         _IOW(PX_IOC_MAGIC,13, Px_Size)
#define PX_IOCGSIZE         _IOR(PX_IOC_MAGIC,14, Px_Size)

#define PX_IOCAUTOGAIN     _IOW(PX_IOC_MAGIC,15, int)
/* ... more to come: offset, camera, compare, count */

#define PX_IOC_MAXNR 15

/* These two structures are used by ioctl */
typedef struct Px_Gain {
    unsigned char range;
    unsigned char fine;
} Px_Gain;

typedef struct Px_Register {
    unsigned long reg;
    unsigned long val;
} Px_Register;

typedef struct Px_Size {
    unsigned int xsize;
    unsigned int ysize;
    unsigned int firstline;
    unsigned int lastline;
    unsigned int hole;
} Px_Size;

typedef unsigned char Px_Lut[256];

/*
 * The device is described by this structure. It tries to keep
 * data aligned both on the PC and the alpha (long/ptr is 4 or 8, int is 4)
 */

/* FIXME: realign data */
typedef struct Px_Dev {
    char *mapbuffer;          /* map of allocated pages */
    unsigned long dmabuffer;  /* physical address */
    void *vdmabuffer;         /* ioremapped address */
    unsigned long dmahole;    /* hole beween lines */
    unsigned long flags;      /* various flags */

    struct Px_Dev *next;
    volatile __u8 *regs;     /* where register are accessed */
    struct wait_queue *queue;
    unsigned long irq_count;

    unsigned int dmasize;    /* allocated size */
    unsigned int imagesize;  /* used size */
    unsigned int xsize;      /* the pgm node needs such info */
    unsigned int ysize;

    __u32 mode;              /* cache datum in register @0x40 */
    __u16 firstline, lastline;
    __u16 port0, port2;       /* cache bits in bitmask ports */

    int usage;               /* number of open */
    int isup;                /* whether dma is active */

    __u32 phys_add;          /* where registers live */
    __u8 bus;                /* pci bus for this board */
    __u8 device;             /* pci device and function */
    __u8 irq;                /* irq number from pci config */
    __u8 irq_active;         /* currently used. 0 if unused */

    int autogain_step;
    int autogain_now;

    Px_Gain gain;            /* cache the current gain */
    Px_Lut lut;              /* cache the current lut */
    char pgmheader[32];      /* the pgm header is 17 bytes */
    #define PX_HEADERLEN 17  /* remember it.. */
    #define PX_HEADERFMT "P5 %4i %4i 255\n"
} Px_Dev;

#define PX_FLAG_AUTOGAIN 0x00000001  /* automatically adjust camera gain */


#ifdef __KERNEL__ /* the rest isn't meant to be seen by userspace */

#define PX_PORT_MIN 0x30
#define PX_PORT_MAX 0x37

/* Use macros to access regs */

#define PX_READ32(d, r) (readl((d)->regs+(r)))
#define PX_READ16(d, r) (readw((d)->regs+(r)))
#define PX_READ8(d, r)  (readb((d)->regs+(r)))

#define PX_WRITE32(d, v, r) (writel(v,(d)->regs+(r)))
#define PX_WRITE16(d, v, r) (writew(v,(d)->regs+(r)))
#define PX_WRITE8(d, v, r)  (writeb(v,(d)->regs+(r)))


extern inline void px_set_mode(Px_Dev *d, u32 mode)
{
  if (d->mode == mode) return;
  d->mode = mode;
  PX_WRITE32(d, mode, __PX_REG_GRABCTL);
}

extern void px_wait_port(Px_Dev *d);

/*
 * Use inline functions to do easy tasks
 */

/* error checking for port access */

extern inline void px_clrerr(Px_Dev *d)
{
    PX_WRITE32(d, 0x200 | PX_READ32(d, PX_REG_SERCTL), PX_REG_SERCTL);
}

extern inline int px_iserr(Px_Dev *d)
{
    if (PX_READ32(d, PX_REG_SERCTL) & 0x200) {
	px_clrerr(d);
	return 1;
    }
    return 0;
}

/* actual port access */

extern inline void __px_set_port(Px_Dev *d, int port, int val)
{
    PX_WRITE32(d, (val<<8) | (port<<17) | (1<<24), PX_REG_SERDATA);
}

extern inline int px_set_port(Px_Dev *d, int port, int val)
{
    /* errors are assumed to be clear */
    __px_set_port(d, port, val);
    if (px_iserr(d)) {
        px_clrerr(d);
	return -EIO;
    }
    return 0;
}

extern inline int px_set_port_wait(Px_Dev *d, int port, int val)
{
    px_clrerr(d);
    __px_set_port(d,port,val);
    px_wait_port(d);
    if (px_iserr(d)) {
        px_clrerr(d);
        return -EIO;
    }
    return 0;
}

extern inline int px_get_port(Px_Dev *d, int port)
{
    px_clrerr(d);
    PX_WRITE32(d, (port<<17) | (1<<24) | (1<<16), PX_REG_SERDATA);
    px_wait_port(d);
    if (px_iserr(d)) {
        px_clrerr(d);
        return -EIO;
    }
    return 0xff & (PX_READ32(d, PX_REG_SERCTL) >> 24);
}

/* then, most used ports deserve special functions */

extern inline int px_set_control_bit(Px_Dev *d, int bit)
{
    return px_set_port(d,__PX_PORT_CONTROL, d->port0 |= bit);
}

extern inline int px_clr_control_bit(Px_Dev *d, int bit)
{
    return px_set_port(d,__PX_PORT_CONTROL, d->port0 &= ~bit);
}

extern inline int px_set_trigger_bit(Px_Dev *d, int bit)
{
    return px_set_port(d,__PX_PORT_TRIGGER, d->port2 |= bit);
}

extern inline int px_clr_trigger_bit(Px_Dev *d, int bit)
{
    return px_set_port(d,__PX_PORT_TRIGGER, d->port2 &= ~bit);
}

extern inline void px_reset_ports(Px_Dev *d)
{
    udelay(100);
    px_set_port_wait(d, 0, 0); d->port0=0; /* Hmm... this locks up */
    udelay(100);
    px_set_port_wait(d, 2, 0); d->port2=0; /* and this too */
}

extern inline int px_decimate(Px_Dev *d, int decimate)
{
  int p = d->port2 &0x8f;
  d->port2 = (p  |= (decimate << 4));
  return px_set_port(d,__PX_PORT_TRIGGER,p);
}

/*
 * Split minors in two parts
 */
#define TYPE(dev)   (MINOR(dev) >> 4)  /* high nibble */
#define NUM(dev)    (MINOR(dev) & 0xf) /* low  nibble */

enum px_types {
    PX_BIN = 0,   /* binary grabber */
    PX_PGM,       /* pgm grabber */
    PX_CTL        /* no grab: only control operations */
};

extern struct file_operations px_fops[];   /* one fops for type */

/*
 * The different configurable parameters
 */

#define    PX_MAJOR 60   /* not dynamic major by default -- FIXME -- */
extern int px_major;

extern Px_Dev *px_devices;

/*
 * Prototypes for shared functions
 */


#ifdef PX_DEBUG 
#  if LINUX_VERSION_CODE > VERSION_CODE(1,99,3)
#    define PX_USE_PROC
#  endif
#endif

#ifndef min
#  define min(a,b) ((a)<(b) ? (a) : (b))
#endif

#else /* not __KERNEL__ */

extern unsigned char px_tmp;
extern int px_res;
extern Px_Register px_reg;

/* if someone doesn't want to link the library... */
#define PX_VARIABLES static unsigned char px_tmp; \
                     static int px_res; \
                     static Px_Register px_reg;

/*
 * These macros are wrappers to ioctl()
 * They use gcc extensions to the C language, but every Linux runs gcc
 */

/* Reset. Hardreset not wrapped, as it is only a debugging tool */

#define Px_Reset(fd) (ioctl((fd), PX_IOCRESET))

/* command-based information */

#define Px_Command(fd,cmd,arg) (px_reg.reg = (cmd), \
				px_reg.val = (arg), \
				ioctl((fd),PX_IOCCMD,&px_reg) ? -1 : \
				px_reg.val)
#define Px_ProtKey(fd)   (Px_Command((fd),PX_CMD_PROTECTION, 0))
#define Px_Revision(fd)  (Px_Command((fd),PX_CMD_REVISION, 0))
#define Px_Status(fd)    (Px_Command((fd),PX_CMD_STATUS, 0))

#define PX_NTSC_VIDEO 0x55 /* two random numbers */
#define PX_PAL_VIDEO  0xaa

#define Px_VideoType(fd) (px_res = Px_Status(fd), px_res < 0 ? -1 : \
			  (px_res&PX_BIT_PAL ? PX_PAL_VIDEO : PX_NTSC_VIDEO))
#define Px_PalVideo(fd)  (Px_VideoType(fd) == PX_PAL_VIDEO)
#define Px_NtscVideo(fd) (Px_VideoType(fd) == PX_NTSC_VIDEO)

#define Px_HasRam(fd)    (px_res = Px_Status(fd), px_res < 0 ? -1 : \
			  (px_res&PX_BIT_NORAM ? 0 : 1))

/* registers and ports: FIXME: need interface for setbit and grabctl */

#define Px_SetReg(reg, arg) (px_reg.reg = (reg), \
			     px_reg.val = (arg), \
			     ioctl((fd),PX_IOCSREG,&px_reg))
#define Px_GetReg(reg) (px_reg.reg = (reg), \
			ioctl((fd),PX_IOCGREG,&px_reg) ? -1 : \
			x_reg.val)

#define Px_SetPort(port, arg) (px_reg.reg = (port), \
			       px_reg.val = (arg), \
			       ioctl((fd),PX_IOCSPORT,&px_reg))
#define Px_GetPort(fd,port) (px_reg.reg = (port), \
			     ioctl((fd),PX_IOCGPORT,&px_reg) ? -1 : \
			     px_reg.val)

#define Px_HaveVideo(fd) (px_res = Px_GetPort((fd),PX_PORT_STATUS), \
			  px_res < 0 ? px_res : !!(px_res & PX_BIT_HAVE_VIDEO))

#define Px_SetGain(fd,ptr) (ioctl((fd),PX_IOCSGAIN,(ptr)))
#define Px_GetGain(fd,ptr) (ioctl((fd),PX_IOCGGAIN,(ptr)))

#define Px_SetLut(fd,ptr) (ioctl((fd),PX_IOCSLUT,(ptr)))
#define Px_GetLut(fd,ptr) (ioctl((fd),PX_IOCGLUT,(ptr)))

#define Px_GetDev(fd,ptr) (ioctl((fd),PX_IOCGDEV,(ptr)))

#define Px_GetSize(fd,ptr) (ioctl((fd),PX_IOCGSIZE,(ptr)))
#define Px_SetSize(fd,ptr) (ioctl((fd),PX_IOCSSIZE,(ptr)))

#define Px_AutoGain(fd,flag) (px_res = (flag), \
			      ioctl((fd), PX_IOCAUTOGAIN, &px_res))

#endif /* __KERNEL__ */

#endif /* _LINUX_PX500_H_ */
